use embedded_graphics::{
    geometry::Point,
    mono_font::{ascii::FONT_10X20, MonoTextStyleBuilder},
    pixelcolor::{Rgb565, RgbColor},
    prelude::DrawTarget,
    text::{Baseline, Text},
    Drawable,
};
use fugit::Duration;
use heapless::{String, Vec};
use rp2040_hal::timer::Instant;

use crate::{
    app::{monotonics, MySpi, HEIGHT, WIDTH},
    gui::{
        chat::Chat, chat_list::ChatList, contact_view::ContactView, selector::Selector,
        settings::SettingsView, status::StatusBar, Element,
    },
    keyboard::KeyCode,
    radio::RadioController,
    storage::{ContactIter, Storage, MAX_CONTACTS, MAX_CONTACT_NAME_LENGTH},
    voltage::VoltMeter,
};

const TX_QUEUE_SIZE: usize = 64;
const BACKOFF_INITIAL: u32 = 3; // seconds
const BACKOFF_FACTOR: u32 = 1;
const RX_LIGHT_TIME_MS: u64 = 100; // milliseconds

pub enum View {
    MainMenu(Selector<&'static str, [&'static str; 3]>),
    Contacts(
        Selector<
            String<MAX_CONTACT_NAME_LENGTH>,
            Vec<String<MAX_CONTACT_NAME_LENGTH>, MAX_CONTACTS>,
        >,
    ),
    ContactOptions(Selector<&'static str, [&'static str; 3]>, usize),
    #[allow(clippy::enum_variant_names)]
    ContactView(ContactView, usize, usize),
    ChatList(ChatList),
    Chat(Chat),
    Settings(SettingsView),
}

impl View {
    pub fn new_main_menu(id: usize) -> Self {
        let menu = Selector::new(["Contacts", "Messages", "Settings"], id);
        Self::MainMenu(menu)
    }

    pub fn new_contacts(id: usize, contacts: &ContactIter) -> Self {
        let mut contact_names: Vec<_, MAX_CONTACTS> = contacts.map(|c| c.name.clone()).collect();

        // If we run out of space, we automatically hide the "Add" button
        let _ = contact_names.push(String::try_from("Add New").unwrap());

        Self::Contacts(Selector::new(contact_names, id))
    }

    pub fn new_contact_options(contact_id: usize, selected_id: usize) -> Self {
        let opts = Selector::new(["View", "Edit", "Message"], selected_id);
        Self::ContactOptions(opts, contact_id)
    }

    pub fn from_main_menu_id(id: usize, contacts: &ContactIter, storage: &Storage) -> Option<Self> {
        match id {
            0 => Some(Self::new_contacts(0, contacts)),
            1 => Some(Self::ChatList(ChatList::new(storage, 0))),
            2 => Some(Self::Settings(SettingsView::new(storage))),
            _ => unreachable!(),
        }
    }

    pub fn to_main_menu_id(&self, storage: &mut Storage) -> Option<usize> {
        let ret = match self {
            View::Contacts(_) => 0,
            View::ChatList(_) => 1,
            View::Settings(s) => {
                if !s.exit(storage) {
                    // validation failed
                    // don't exit; let the user correct it
                    return None;
                }

                2
            }
            _ => unreachable!(),
        };

        Some(ret)
    }

    pub fn from_contact_options_id(
        id: usize,
        contact_id: usize,
        storage: &Storage,
    ) -> Option<Self> {
        let mut contacts = storage.contacts();

        match id {
            0 => {
                let contact = contacts
                    .nth(contact_id)
                    .unwrap_or_else(|| storage.new_contact());
                Some(View::ContactView(
                    ContactView::new(contact, false),
                    contact_id,
                    id,
                ))
            }

            1 => {
                let contact = contacts
                    .nth(contact_id)
                    .unwrap_or_else(|| storage.new_contact());
                Some(View::ContactView(
                    ContactView::new(contact, true),
                    contact_id,
                    id,
                ))
            }

            2 => {
                let contact = contacts
                    .nth(contact_id)
                    .unwrap_or_else(|| storage.new_contact());

                Some(View::Chat(Chat::new(contact.callsign, contact.ssid)))
            }

            _ => unreachable!(),
        }
    }

    fn render<E, DT: DrawTarget<Color = Rgb565, Error = E>>(
        &mut self,
        dt: &mut DT,
        storage: &Storage,
    ) -> Result<(), E> {
        match self {
            View::MainMenu(m) => m.render(dt),
            View::Contacts(c) => c.render(dt),
            View::ContactOptions(c, _) => c.render(dt),
            View::ContactView(c, _, _) => c.render(dt),
            View::ChatList(c) => c.render(dt, storage),
            View::Chat(c) => c.render(dt, storage),
            View::Settings(s) => s.render(dt),
        }
    }

    fn key_push(&mut self, k: KeyCode, storage: &mut Storage) -> Option<u16> {
        match self {
            View::MainMenu(m) => m.key_push(k),
            View::Contacts(c) => c.key_push(k),
            View::ContactOptions(c, _) => c.key_push(k),
            View::ContactView(c, _, _) => {
                c.key_push(k);
            }
            View::ChatList(c) => c.key_push(k),
            View::Chat(c) => return c.key_push(k, storage),
            View::Settings(s) => s.key_push(k),
        }

        None
    }

    fn touchpad_scroll(&mut self, x: i8, y: i8, brightness: &mut u8) {
        match self {
            View::MainMenu(m) => m.touchpad_scroll(x, y),
            View::Contacts(c) => c.touchpad_scroll(x, y),
            View::ContactOptions(c, _) => c.touchpad_scroll(x, y),
            View::ContactView(c, _, _) => c.touchpad_scroll(x, y),
            View::ChatList(c) => c.touchpad_scroll(x, y),
            View::Chat(c) => c.touchpad_scroll(x, y),
            View::Settings(s) => s.touchpad_scroll(x, y, brightness),
        }
    }
}

impl Default for View {
    fn default() -> Self {
        Self::new_main_menu(0)
    }
}

#[derive(Copy, Clone)]
pub struct TxItem {
    id: u16,
    next_send_at: Instant,
}

pub struct Controller {
    status: StatusBar,
    cur_view: View,
    storage: Storage,
    radio: RadioController,
    tx_queue: Vec<TxItem, TX_QUEUE_SIZE>,
    green_led: Option<Instant>,
}

impl Controller {
    pub fn new(storage: Storage, volt_meter: VoltMeter, radio: RadioController) -> Self {
        Self {
            status: StatusBar::new(volt_meter),
            cur_view: View::default(),
            storage,
            radio,
            tx_queue: Vec::new(),
            green_led: None,
        }
    }

    pub fn tick(&mut self, spi: &mut MySpi) {
        if !self.radio.is_busy_rxing(spi) {
            let now = monotonics::now();
            let msg = self
                .tx_queue
                .iter()
                .enumerate()
                .find(|(_, x)| x.next_send_at <= now);

            if let Some((id, &msg)) = msg {
                let mut message = self.storage.get_message(msg.id).unwrap();

                if self
                    .radio
                    .send_message(&self.storage, &message, msg.id, spi)
                    .is_none()
                {
                    message.send_attempts = 255;
                    self.tx_queue.remove(id);
                } else {
                    // re-enqueue message in tx queue w/ exponential backoff
                    // only gets removed fr when we get an ack (or we run out of retries)
                    message.send_attempts += 1;

                    if message.failed() {
                        self.tx_queue.remove(id);
                    } else {
                        let delay_secs =
                            BACKOFF_INITIAL * BACKOFF_FACTOR.pow(message.send_attempts.into());

                        self.tx_queue[id].next_send_at = monotonics::now()
                            + Duration::<u64, 1, 1000000>::secs(delay_secs.into());
                    }
                }

                message.write(self.storage.flash_mut(), msg.id.into());
            }
        }

        let (for_us, resp) = self.radio.tick(spi, &mut self.storage);
        if let Some(msg_id) = resp {
            self.refresh_rx_led();

            while let Some((tx_id, _)) = self
                .tx_queue
                .iter()
                .enumerate()
                .find(|(_, tx)| tx.id == msg_id)
            {
                self.tx_queue.swap_remove(tx_id);
            }
        }

        if for_us {
            self.refresh_rx_led();
        }
    }

    pub fn red_led_on(&self) -> bool {
        self.radio.is_transmitting()
    }

    pub fn green_led_on(&self) -> bool {
        match self.green_led {
            Some(x) => monotonics::now() < x,
            None => false,
        }
    }

    fn refresh_rx_led(&mut self) {
        self.green_led =
            Some(monotonics::now() + Duration::<u64, 1, 1000000>::millis(RX_LIGHT_TIME_MS));
    }

    fn handle_message_tx(&mut self, id: u16) {
        let tx = TxItem {
            id,
            next_send_at: monotonics::now(),
        };

        let enqueued = self.tx_queue.push(tx).is_ok();

        if !enqueued {
            let worst = self
                .tx_queue
                .iter()
                .enumerate()
                .filter_map(|(id, tx)| Some((id, self.storage.get_message(tx.id)?)))
                .max_by_key(|(_, msg)| msg.send_attempts)
                .unwrap() // safe to unwrap - if we're full we definitely have a message
                .0;

            self.tx_queue[worst] = tx;
        }
    }

    pub fn render<E, DT: DrawTarget<Color = Rgb565, Error = E>>(
        &mut self,
        dt: &mut DT,
    ) -> Result<(), E> {
        if self.radio.errored() {
            dt.clear(Rgb565::new(16, 0, 0))?;

            self.status.render(dt)?;

            let label_style = MonoTextStyleBuilder::new()
                .font(&FONT_10X20)
                .background_color(Rgb565::WHITE)
                .text_color(Rgb565::BLACK)
                .build();

            Text::with_baseline(
                "Error communicating with\nradio module",
                Point::new(
                    (WIDTH / 2 - 120).try_into().unwrap(),
                    (HEIGHT / 2).try_into().unwrap(),
                ),
                label_style,
                Baseline::Middle,
            )
            .draw(dt)?;
        } else {
            dt.clear(Rgb565::new(0, 32, 16))?;

            self.cur_view.render(dt, &self.storage)?;

            self.status.render(dt)?;
        }

        Ok(())
    }

    pub fn key_push(&mut self, k: KeyCode) {
        match (&mut self.cur_view, k) {
            (View::MainMenu(m), KeyCode::Touchpad) => {
                if let Some(new_view) =
                    View::from_main_menu_id(m.selected(), &self.storage.contacts(), &self.storage)
                {
                    self.cur_view = new_view;
                }
            }

            (View::Contacts(c), KeyCode::Touchpad) => {
                if c.selected() == self.storage.contacts_len() {
                    // Go direct to edit page
                    self.cur_view =
                        View::from_contact_options_id(1, c.selected(), &self.storage).unwrap();
                } else {
                    self.cur_view = View::new_contact_options(c.selected(), 0);
                }
            }

            (View::ContactOptions(c, contact_id), KeyCode::Touchpad) => {
                if let Some(new_view) =
                    View::from_contact_options_id(c.selected(), *contact_id, &self.storage)
                {
                    self.cur_view = new_view;
                }
            }

            (View::ChatList(cl), KeyCode::Touchpad) => {
                if let Some((callsign, ssid)) = self.storage.message_addressees().nth(cl.selected())
                {
                    self.cur_view = View::Chat(Chat::new(callsign, ssid));
                }
            }

            (View::MainMenu(_), KeyCode::Back) => {}

            (View::ContactOptions(_, id), KeyCode::Back) => {
                self.cur_view = View::new_contacts(*id, &self.storage.contacts());
            }

            (View::ContactView(cv, contact_id, selected_id), KeyCode::Back) => {
                // TODO only need to call this if we actually edited
                self.storage.save_contact(cv.contact());

                self.cur_view = View::new_contact_options(*contact_id, *selected_id);
            }

            (View::Chat(cl), KeyCode::Back) => {
                let selected_id = self
                    .storage
                    .message_addressees()
                    .enumerate()
                    .find(|(_, (call, ssid))| call == cl.callsign() && *ssid == cl.ssid())
                    .map(|(id, _)| id)
                    .unwrap_or(0);
                self.cur_view = View::ChatList(ChatList::new(&self.storage, selected_id));
            }

            (_, KeyCode::Back) => {
                if let Some(id) = self.cur_view.to_main_menu_id(&mut self.storage) {
                    self.cur_view = View::new_main_menu(id);
                }
            }

            (View::ContactView(cv, contact_id, _), k) => {
                if cv.key_push(k) {
                    // if statement to handle edge case where contact is created, then deleted, without getting saved
                    if *contact_id < self.storage.contacts_len() {
                        self.storage.delete_contact(*contact_id);
                    }

                    self.cur_view = View::new_contacts(*contact_id, &self.storage.contacts());
                }
            }

            (_, k) => {
                if let Some(id) = self.cur_view.key_push(k, &mut self.storage) {
                    self.handle_message_tx(id)
                }
            }
        }
    }

    pub fn touchpad_scroll(&mut self, x: i8, y: i8, brightness: &mut u8) {
        self.cur_view.touchpad_scroll(x, y, brightness);
    }
}
