use embedded_hal::{
    blocking::{delay::DelayUs, spi::Transfer},
    digital::v2::OutputPin,
    spi::FullDuplex,
};
use rp2040_hal::{
    dma::{single_buffer, Channel, CH0},
    gpio::{
        bank0::{Gpio13, Gpio16},
        FunctionSio, Pin, PullDown, SioOutput,
    },
};

use crate::app::{MySpi, HEIGHT, WIDTH};

type Cs = Pin<Gpio13, FunctionSio<SioOutput>, PullDown>;
type Dc = Pin<Gpio16, FunctionSio<SioOutput>, PullDown>;

type MyDma = Channel<CH0>;

enum Command {
    SoftReset = 0x01,
    ExitSleepMode = 0x11,
    EnterNormalMode = 0x13,
    SetUninvertedMode = 0x20,
    SetDisplayOn = 0x29,
    WriteMemoryStart = 0x2C,
    SetScrollArea = 0x33,
    SetAddressMode = 0x36,
    SetPixelFormat = 0x3A,
}

pub struct St7789 {
    cs: Cs,
    dc: Dc,
    dma: Option<MyDma>,
}

impl St7789 {
    pub fn new<D: DelayUs<u32>>(spi: &mut MySpi, cs: Cs, dc: Dc, dma: MyDma, mut delay: D) -> Self {
        let mut s = Self {
            cs,
            dc,
            dma: Some(dma),
        };

        s.send_command(spi, Command::SoftReset, []);

        delay.delay_us(150_000);

        s.send_command(spi, Command::ExitSleepMode, []);

        delay.delay_us(10_000);

        // set framebuffer/scroll stuff
        let vsa: u16 = WIDTH.max(HEIGHT).try_into().unwrap();
        let vsa_bytes = vsa.to_be_bytes();
        s.send_command(
            spi,
            Command::SetScrollArea,
            [0, 0, vsa_bytes[0], vsa_bytes[1], 0, 0],
        );

        // inverted landscape, RGB
        s.send_command(spi, Command::SetAddressMode, [0b1000_0000]);

        // normal mode
        s.send_command(spi, Command::SetUninvertedMode, []);

        // 16 bits per pixel
        s.send_command(spi, Command::SetPixelFormat, [0b101]);

        delay.delay_us(10_000);

        s.send_command(spi, Command::EnterNormalMode, []);

        delay.delay_us(10_000);

        s.send_command(spi, Command::SetDisplayOn, []);

        delay.delay_us(120_000);

        s
    }

    /// Panics if opt_spi is None
    pub fn write_framebuffer(
        &mut self,
        opt_spi: &mut Option<MySpi>,
        buffer: &'static mut [u8; WIDTH * HEIGHT * 2],
    ) -> &'static mut [u8; WIDTH * HEIGHT * 2] {
        self.send_command(opt_spi.as_mut().unwrap(), Command::WriteMemoryStart, []);

        self.cs.set_low().unwrap();

        let transfer =
            single_buffer::Config::new(self.dma.take().unwrap(), buffer, opt_spi.take().unwrap())
                .start();

        let (channel, buffer, mut spi) = transfer.wait();
        self.dma = Some(channel);

        while spi.is_busy() {}

        while spi.read().is_ok() {}

        *opt_spi = Some(spi);

        self.cs.set_high().unwrap();
        self.dc.set_low().unwrap();

        buffer
    }

    fn send_command<const N: usize>(
        &mut self,
        spi: &mut MySpi,
        command: Command,
        mut data: [u8; N],
    ) {
        self.dc.set_low().unwrap();

        self.cs.set_low().unwrap();
        spi.transfer(&mut [command as u8]).unwrap();
        self.cs.set_high().unwrap();

        self.dc.set_high().unwrap();

        if !data.is_empty() {
            self.cs.set_low().unwrap();
            spi.transfer(&mut data).unwrap();
            self.cs.set_high().unwrap();
        }
    }
}
