use embedded_graphics::{
    geometry::{Point, Size},
    mono_font::{ascii::FONT_10X20, MonoTextStyle},
    pixelcolor::{Rgb565, RgbColor},
    primitives::{Primitive, PrimitiveStyleBuilder, Rectangle},
    text::renderer::TextRenderer,
    Drawable,
};
use embedded_text::{style::TextBoxStyleBuilder, TextBox};
use heapless::String;

use crate::{
    app::{HEIGHT, WIDTH},
    keyboard::KeyCode,
    storage::{
        Message, MessageDirection, Storage, MAX_CONTACT_CALLSIGN_LENGTH, MAX_MESSAGE_BODY_LENGTH,
    },
};

use super::{textbox::TextBox as MyTextBox, Element};

const MESSAGE_BOX_WIDTH: u32 = 200;
const MESSAGE_BOX_HEIGHT: u32 = 10_000; // Unlimited

pub struct Chat {
    callsign: String<MAX_CONTACT_CALLSIGN_LENGTH>,
    ssid: u8,
    textbox: MyTextBox<MAX_MESSAGE_BODY_LENGTH>,
}

impl Chat {
    pub fn new(callsign: String<MAX_CONTACT_CALLSIGN_LENGTH>, ssid: u8) -> Self {
        Self {
            callsign,
            ssid,
            textbox: MyTextBox::new(1, i32::try_from(HEIGHT).unwrap() - 24, 31, 1, 4),
        }
    }

    pub fn callsign(&self) -> &str {
        &self.callsign
    }

    pub fn ssid(&self) -> u8 {
        self.ssid
    }

    pub fn render<
        E,
        DT: embedded_graphics::prelude::DrawTarget<
            Color = embedded_graphics::pixelcolor::Rgb565,
            Error = E,
        >,
    >(
        &mut self,
        dt: &mut DT,
        storage: &Storage,
    ) -> Result<(), E> {
        let mut y =
            i32::try_from(HEIGHT).unwrap() - i32::try_from(self.textbox.height()).unwrap() - 8;
        for msg in storage.messages(&self.callsign, self.ssid).rev() {
            let bounds = Rectangle::new(
                Point::new(4, 0),
                Size::new(MESSAGE_BOX_WIDTH, MESSAGE_BOX_HEIGHT),
            );

            let (bg, fg) = if msg.direction() == MessageDirection::From {
                (Rgb565::WHITE, Rgb565::BLACK)
            } else if msg.received {
                (Rgb565::GREEN, Rgb565::BLACK)
            } else if msg.failed() {
                (Rgb565::RED, Rgb565::WHITE)
            } else {
                (Rgb565::BLUE, Rgb565::WHITE)
            };

            let textbox_style = TextBoxStyleBuilder::new().build();
            let character_style = MonoTextStyle::new(&FONT_10X20, fg);

            let mut text_box =
                TextBox::with_textbox_style(msg.content(), bounds, character_style, textbox_style);

            let mut actual_bounds = calculate_bounding_box(&text_box);
            if msg.direction() == MessageDirection::To {
                let x = i32::try_from(WIDTH).unwrap()
                    - i32::try_from(actual_bounds.size.width).unwrap()
                    - 4;

                text_box.bounds.top_left.x = x;
                actual_bounds.top_left.x = x;
            }

            let rect_style = PrimitiveStyleBuilder::new().fill_color(bg).build();

            y -= i32::try_from(actual_bounds.size.height).unwrap() + 4;
            text_box.bounds.top_left.y = y;
            actual_bounds.top_left.y = y;

            actual_bounds.into_styled(rect_style).draw(dt)?;

            text_box.draw(dt)?;

            if y < 0 {
                break;
            }
        }

        self.textbox.render(dt)?;

        Ok(())
    }

    pub fn key_push(&mut self, k: KeyCode, storage: &mut Storage) -> Option<u16> {
        if k == KeyCode::Enter {
            if !self.textbox.get_text().trim().is_empty() {
                let text = self.textbox.clear();
                let message =
                    Message::new(self.callsign.clone(), self.ssid, text, MessageDirection::To);

                let id = storage.push_message(message);

                Some(id)
            } else {
                None
            }
        } else {
            self.textbox.key_push(k);

            None
        }
    }

    pub fn touchpad_scroll(&mut self, x: i8, y: i8) {
        self.textbox.touchpad_scroll(x, y);
    }
}

fn calculate_bounding_box<S: TextRenderer>(tb: &TextBox<'_, S>) -> Rectangle {
    const CHAR_WIDTH: u32 = 10;
    const CHAR_HEIGHT: u32 = 20;

    let mut x = 0;
    let mut max_x = 0;
    let mut y = 1;
    for c in tb.text.chars() {
        if c == '\n' || CHAR_WIDTH * x >= tb.bounds.size.width {
            x = 0;
            y += 1;
        } else {
            x += 1;

            if x > max_x {
                max_x = x;
            }
        }
    }

    Rectangle::new(
        tb.bounds.top_left,
        Size::new(max_x * CHAR_WIDTH, y * CHAR_HEIGHT),
    )
}
