use ham_cats::{
    buffer::Buffer,
    packet::Packet,
    whisker::{Destination, Identification, NodeInfoBuilder, Route},
};
use rf4463::{config, Rf4463};
use rp2040_hal::{
    gpio::{
        bank0::{Gpio19, Gpio21},
        FunctionSio, Pin, PullDown, SioOutput,
    },
    Timer,
};

use crate::{
    app::MySpi,
    storage::{Message, MessageDirection, Storage, MAX_MESSAGE_BODY_LENGTH},
};

type SdnPin = Pin<Gpio19, FunctionSio<SioOutput>, PullDown>;
type CsPin = Pin<Gpio21, FunctionSio<SioOutput>, PullDown>;

type MyRf4463 = Rf4463<SdnPin, CsPin, Timer>;

pub const MAX_PACKET_LEN: usize = 8191;
const HARDWARE_ID: u16 = 0x7c87;
const SOFTWARE_ID: u8 = 0;
const MAX_HOPS: u8 = 3;

pub struct RadioController {
    radio: Option<MyRf4463>,
    buf: &'static mut [u8; MAX_PACKET_LEN],
}

impl RadioController {
    pub fn new(
        storage: &Storage,
        spi: &mut MySpi,
        sdn: SdnPin,
        cs: CsPin,
        delay: Timer,
        buf: &'static mut [u8; MAX_PACKET_LEN],
    ) -> Self {
        let mut config = config::RADIO_CONFIG_CATS;

        let radio = Rf4463::new(spi, sdn, cs, delay, &mut config).ok();

        let mut s = Self { radio, buf };

        Self::radio_handler(&mut s.radio, |rad| {
            rad.set_frequency(spi, storage.settings().frequency())
        });

        s
    }

    // Returns (message was destined for us, the message id of the message that was ack'd, if any)
    // (Yes, I know the fact that we're doing that is kind of spaghetti)
    pub fn tick(&mut self, spi: &mut MySpi, storage: &mut Storage) -> (bool, Option<u16>) {
        Self::radio_handler_opt(&mut self.radio, |rad| {
            if rad.is_busy_txing() {
                rad.interrupt(spi, None, Some(self.buf)).ok()?;
            } else {
                rad.interrupt(spi, Some(self.buf), None).ok()?;
            }

            Some(())
        });

        if let Some(ref mut rad) = self.radio {
            let mut buf = [0; MAX_PACKET_LEN];

            let Ok(data) = rad.finish_rx(spi, &mut self.buf[..]) else {
                self.radio = None;
                return (false, None);
            };

            if let Some(data) = data {
                if let Ok(packet) = Packet::fully_decode(data.data(), &mut buf) {
                    let for_us = self.handle_rx(spi, storage, &packet).is_some();

                    // not very DRY that this is duplicated TODO
                    Self::radio_handler_opt(&mut self.radio, |rad| {
                        if rad.is_idle() {
                            rad.start_rx(spi, None, false).ok()?;
                        }

                        Some(())
                    });

                    return (for_us, self.handle_ack(storage, &packet));
                }
            }
        }

        Self::radio_handler_opt(&mut self.radio, |rad| {
            if rad.is_idle() {
                rad.start_rx(spi, None, false).ok()?;
            }

            Some(())
        });

        (false, None)
    }

    pub fn is_busy_rxing(&mut self, spi: &mut MySpi) -> bool {
        Self::radio_handler(&mut self.radio, |rad| rad.is_busy_rxing(spi)).unwrap_or(true)
    }

    pub fn send_message(
        &mut self,
        storage: &Storage,
        message: &Message,
        id: u16,
        spi: &mut MySpi,
    ) -> Option<()> {
        let settings = storage.settings();

        if !settings.allowed_callsign() {
            return None;
        }

        // Construct CATS packet
        let mut buf = [0; MAX_PACKET_LEN];
        let mut pkt = Packet::new(&mut buf);
        pkt.add_identification(Identification {
            callsign: settings.callsign().try_into().unwrap(),
            ssid: settings.ssid(),
            icon: settings.icon(),
        })
        .ok()?;

        // TODO could do battery voltage here
        pkt.add_node_info(
            NodeInfoBuilder::default()
                .hardware_id(HARDWARE_ID)
                .software_id(SOFTWARE_ID)
                .build(),
        )
        .ok()?;

        pkt.add_destination(Destination::new(
            false,
            (id & 127).try_into().unwrap(),
            message.callsign(),
            message.ssid(),
        )?)
        .ok()?;

        pkt.add_route(Route::new(MAX_HOPS)).ok()?;

        pkt.add_comment(message.content()).ok()?;

        let mut out = Buffer::new_empty(self.buf);
        pkt.fully_encode(&mut out).ok()?;

        Self::radio_handler(&mut self.radio, |rad| rad.start_tx(spi, &out));

        Some(())
    }

    pub fn errored(&self) -> bool {
        self.radio.is_none()
    }

    pub fn is_transmitting(&self) -> bool {
        self.radio
            .as_ref()
            .map(|r| r.is_busy_txing())
            .unwrap_or(false)
    }

    fn handle_ack(
        &mut self,
        storage: &mut Storage,
        packet: &Packet<MAX_PACKET_LEN>,
    ) -> Option<u16> {
        let settings = storage.settings();

        let dest = packet.destination_iter().find(|d| {
            d.callsign() == settings.callsign() && d.ssid() == settings.ssid() && d.is_ack()
        })?;

        let ident = packet.identification()?;

        let ack_num = dest.ack_num();

        let mut i = ack_num.into();
        while i < storage.messages_len() {
            if let Some(mut msg) = storage.get_message(i) {
                if msg.callsign() == &ident.callsign && msg.ssid() == ident.ssid {
                    // it's a match

                    msg.received = true;
                    msg.write(storage.flash_mut(), i.into());

                    // TODO don't return this. Just check if the message has been received or not
                    return Some(i);
                }
            }

            // only check messages where their id would generate the given ack num
            i += 128;
        }

        None
    }

    fn handle_rx(
        &mut self,
        spi: &mut MySpi,
        storage: &mut Storage,
        packet: &Packet<MAX_PACKET_LEN>,
    ) -> Option<()> {
        let settings = storage.settings();

        // filter to only packets destined for us
        let dest = packet.destination_iter().find(|d| {
            d.callsign() == settings.callsign() && d.ssid() == settings.ssid() && !d.is_ack()
        })?;

        let from = packet.identification()?;
        let mut comment_buf = [0; MAX_MESSAGE_BODY_LENGTH];
        let comment = packet.comment(&mut comment_buf).ok()?;

        if dest.ack_num() > 0 {
            // we *just* received a packet so we know we're safe to immediately switch to tx
            self.send_ack(storage, spi, &from.callsign, from.ssid, dest.ack_num());
        }

        for msg_id in 0..storage.messages_len() {
            if let Some(msg) = storage.get_message(msg_id) {
                if msg.callsign() == &from.callsign
                    && msg.ssid() == from.ssid
                    && msg.ack_number == dest.ack_num()
                    && msg.content() == comment
                    && msg.direction() == MessageDirection::From
                {
                    // duplicate
                    return None;
                }
            }
        }

        let mut msg = Message::new(
            from.callsign.as_str().try_into().ok()?,
            from.ssid,
            comment.try_into().unwrap(), // safe to unwrap - we know it'll fit
            MessageDirection::From,
        );

        msg.ack_number = dest.ack_num();

        storage.push_message(msg);

        Some(())
    }

    fn send_ack(
        &mut self,
        storage: &Storage,
        spi: &mut MySpi,
        to_call: &str,
        to_ssid: u8,
        ack_num: u8,
    ) -> Option<()> {
        let settings = storage.settings();
        if !settings.allowed_callsign() {
            return None;
        }

        // TODO can probably smallen this
        let mut buf = [0; MAX_PACKET_LEN];
        let mut pkt = Packet::new(&mut buf);
        pkt.add_identification(Identification::new(
            settings.callsign(),
            settings.ssid(),
            settings.icon(),
        )?)
        .unwrap();
        pkt.add_route(Route::new(MAX_HOPS)).unwrap();
        pkt.add_destination(Destination::new(true, ack_num, to_call, to_ssid)?)
            .unwrap();
        let mut out = Buffer::new_empty(self.buf);
        pkt.fully_encode(&mut out).ok()?;

        Self::radio_handler(&mut self.radio, |rad| rad.start_tx(spi, &out));

        Some(())
    }

    fn radio_handler<T, E, F: FnMut(&mut MyRf4463) -> Result<T, E>>(
        radio: &mut Option<MyRf4463>,
        mut f: F,
    ) -> Option<T> {
        if let Some(ref mut rad) = radio {
            match f(rad) {
                Ok(x) => return Some(x),
                Err(_) => {
                    *radio = None;
                }
            }
        }

        None
    }

    fn radio_handler_opt<T, F: FnMut(&mut MyRf4463) -> Option<T>>(
        radio: &mut Option<MyRf4463>,
        mut f: F,
    ) -> Option<T> {
        if let Some(ref mut rad) = radio {
            match f(rad) {
                Some(x) => return Some(x),
                None => {
                    *radio = None;
                }
            }
        }

        None
    }
}
