use rp2040_flash::flash;
use sectorize::SectorizeIter;

use core::slice::from_raw_parts;

const SECTOR_SIZE: usize = 4096;
const FLASH_START: usize = 0x10000000;
// First 6MB is for code
pub const DATA_START: usize = FLASH_START + 6144 * 1024;
// 2MB for data
const DATA_END: usize = DATA_START + 2048 * 1024;

pub struct FlashManager {
    // Prevent instantiating a FlashManager without going through new()
    _a: (),
}

impl FlashManager {
    pub(crate) fn new() -> Self {
        Self { _a: () }
    }

    pub(crate) fn write_data(&mut self, start_addr: usize, buf: &[u8]) {
        // rprintln!("Write {:x?} to {:x}", buf, start_addr);

        for s in SectorizeIter::new(SECTOR_SIZE, start_addr, buf.len()) {
            let mut existing = self.read_slice(s.sector_index * SECTOR_SIZE);

            /* rprintln!(
                    "Sector {}: {}..{} <- {}..{}",
                    s.sector_index,
                    s.sector_start,
                    s.sector_end,
                    s.input_start,
                    s.input_end
            ); */

            // only write if there's a difference
            if existing[s.sector_start..s.sector_end] != buf[s.input_start..s.input_end] {
                existing[s.sector_start..s.sector_end]
                    .copy_from_slice(&buf[s.input_start..s.input_end]);
                self.write_sector(s.sector_index * SECTOR_SIZE, &existing);
            }
        }
    }

    pub(crate) fn read_slice<const N: usize>(&self, addr: usize) -> [u8; N] {
        assert!(addr >= DATA_START);
        assert!(addr + N < DATA_END);

        let sector = unsafe { from_raw_parts(addr as *mut u8, N) };

        let mut buf = [0; N];

        buf.copy_from_slice(sector);

        buf
    }

    fn write_sector(&mut self, start_addr: usize, buf: &[u8; 4096]) {
        let addr: u32 = (start_addr - FLASH_START).try_into().unwrap();
        let sector_size = u32::try_from(SECTOR_SIZE).unwrap();
        assert!(addr % sector_size == 0);

        cortex_m::interrupt::free(|_| unsafe {
            flash::flash_range_erase_and_program(addr, buf, true);
        });
    }
}
