use core::{fmt::Write, str::FromStr};

use heapless::{String, Vec};

use crate::storage::X25;

use super::{flash::FlashManager, message::MESSAGE_HEADER_END, MAX_CONTACT_CALLSIGN_LENGTH};

const SETTINGS_START: usize = MESSAGE_HEADER_END;
// 1 byte for ssid
// 2 bytes for icon
// 4 bytes for frequency
// 1 byte for brightness
// 1 byte for callsign length
// 2 bytes for checksum
const SETTINGS_SIZE: usize = 1 + 2 + 4 + 1 + 1 + MAX_CONTACT_CALLSIGN_LENGTH + 2;

#[derive(Clone)]
pub struct Settings {
    callsign: String<MAX_CONTACT_CALLSIGN_LENGTH>,
    ssid: u8,
    icon: u16,
    frequency: u32,
    brightness: u8,
}

impl Settings {
    pub fn new(manager: &FlashManager) -> Self {
        Self::read(manager).unwrap_or_default()
    }

    pub fn allowed_callsign(&self) -> bool {
        !self.callsign.eq_ignore_ascii_case("N0CALL") && !self.callsign.is_empty()
    }

    pub fn set_callsign(&mut self, c: &str) -> Result<(), ()> {
        self.callsign = String::try_from(c)?;

        Ok(())
    }

    pub fn ssid(&self) -> u8 {
        self.ssid
    }

    pub fn set_ssid(&mut self, s: u8) {
        self.ssid = s;
    }

    pub fn icon(&self) -> u16 {
        self.icon
    }

    pub fn set_icon(&mut self, icon: u16) {
        self.icon = icon;
    }

    pub fn frequency(&self) -> u32 {
        self.frequency
    }

    pub fn set_frequency(&mut self, freq: u32) -> Result<(), ()> {
        if !Self::validate_freq(freq) {
            return Err(());
        }

        self.frequency = freq;

        Ok(())
    }

    pub fn brightness(&self) -> u8 {
        self.brightness
    }

    pub fn set_brightness(&mut self, brightness: u8) {
        assert!(brightness <= 100);
        assert!(brightness > 0);

        self.brightness = brightness;
    }

    pub fn callsign(&self) -> &str {
        &self.callsign
    }

    pub fn ssid_string(&self) -> String<3> {
        let mut string = String::new();
        write!(&mut string, "{}", self.ssid).unwrap();

        string
    }

    pub fn icon_string(&self) -> String<5> {
        let mut string = String::new();
        write!(&mut string, "{}", self.icon).unwrap();

        string
    }

    pub fn freq_string(&self) -> String<9> {
        let mut string = String::new();
        write!(&mut string, "{}", self.frequency).unwrap();

        string
    }

    pub fn validate_freq(freq: u32) -> bool {
        (420_000_000..=450_000_000).contains(&freq)
    }

    pub(crate) fn read(manager: &FlashManager) -> Option<Self> {
        let bytes: [_; SETTINGS_SIZE] = manager.read_slice(SETTINGS_START);
        let checksum_expected =
            u16::from_le_bytes([bytes[SETTINGS_SIZE - 2], bytes[SETTINGS_SIZE - 1]]);
        let checksum_actual = X25.checksum(&bytes[..(SETTINGS_SIZE - 2)]);
        if checksum_expected != checksum_actual {
            return None;
        }

        let ssid = bytes[0];
        let icon = u16::from_le_bytes([bytes[1], bytes[2]]);
        let frequency = u32::from_le_bytes(bytes[3..7].try_into().unwrap());
        if !Self::validate_freq(frequency) {
            return None;
        }

        let brightness = bytes[7];
        if !(1..=100).contains(&brightness) {
            return None;
        }

        let callsign_len: usize = bytes[8].into();
        if callsign_len > MAX_CONTACT_CALLSIGN_LENGTH {
            return None;
        }
        let callsign_bytes = &bytes[9..][..callsign_len];
        let callsign = String::from_utf8(Vec::from_slice(callsign_bytes).ok()?).ok()?;

        Some(Self {
            callsign,
            ssid,
            icon,
            frequency,
            brightness,
        })
    }

    pub(crate) fn write(&self, manager: &mut FlashManager) {
        let mut buf = [0; SETTINGS_SIZE];
        buf[0] = self.ssid;
        buf[1..3].copy_from_slice(&self.icon.to_le_bytes());
        buf[3..7].copy_from_slice(&self.frequency.to_le_bytes());
        buf[7] = self.brightness;

        let callsign_bytes = self.callsign.as_bytes();
        buf[8] = callsign_bytes.len().try_into().unwrap();
        buf[9..][..callsign_bytes.len()].copy_from_slice(callsign_bytes);

        let checksum = X25.checksum(&buf[..(SETTINGS_SIZE - 2)]);
        buf[(SETTINGS_SIZE - 2)..].copy_from_slice(&checksum.to_le_bytes());

        manager.write_data(SETTINGS_START, &buf);
    }
}

impl Default for Settings {
    fn default() -> Self {
        Self {
            callsign: String::from_str("N0CALL").unwrap(),
            ssid: 0,
            icon: 0,
            frequency: 430_500_000,
            brightness: 100,
        }
    }
}
